threads / patch / 27148

patchgit-rebase--interactive.sh: preserve-merges fails on merges created with no-ff

Subject: [PATCH] git-rebase--interactive.sh: preserve-merges fails on merges created with no-ff

## tl;dr

6 messages between Apr 21, 2011 and Apr 28, 2011. Diffs are folded; open one to read it.

replies: 5people: 4as markdown or json

Andrew Wong· Apr 21, 2011, 03:38 UTC · lore

'git rebase' uses 'git merge' to preserve merges (-p). This preserves the original merge commit correctly, except when the original merge commit was created by 'git merge --no-ff'. In this case, 'git rebase' will fail to preserve the merge, because during 'git rebase', 'git merge' will simply fast-forward and skip the commit. For example:

               B
              / \
             A---M
            /
    ---o---O---P---Q
If we try to rebase M onto P, we lose the merge commit and this happens:
                 A---B
                /
    ---o---O---P---Q

To correct this, we simply do a "no fast-forward" on all merge commits when rebasing. Since by the time we decided to do a 'git merge' inside 'git rebase', it means there was a merge originally, so 'git merge' should always create a merge commit regardless of what the merge branches look like. This way, when rebase M onto P from the above example, we get:

                   B
                  / \
                 A---M
                /
    ---o---O---P---Q
Signed-off-by: Andrew Wong <andrew.kw.w@gmail.com>
---
 git-rebase--interactive.sh |    2 +-
 1 files changed, 1 insertions(+), 1 deletions(-)
Show changes to git-rebase--interactive.sh +1 −1
diff --git a/git-rebase--interactive.sh b/git-rebase--interactive.sh
index 5873ba4..c308529 100755
--- a/git-rebase--interactive.sh
+++ b/git-rebase--interactive.sh
@@ -339,7 +339,7 @@ pick_one_preserving_merges () {
 			# No point in merging the first parent, that's HEAD
 			new_parents=${new_parents# $first_parent}
 			if ! do_with_author output \
-				git merge $STRATEGY -m "$msg" $new_parents
+				git merge --no-ff $STRATEGY -m "$msg" $new_parents
 			then
 				printf "%s\n" "$msg" > "$GIT_DIR"/MERGE_MSG
 				die_with_patch $sha1 "Error redoing merge $sha1"
-- 
1.7.2.2
Junio C Hamano· Apr 27, 2011, 05:00 UTC · re: Andrew Wong · lore

Re: [PATCH] git-rebase--interactive.sh: preserve-merges fails on merges created with no-ff

Andrew Wong <andrew.w@sohovfx.com> writes:
> Could someone please take a look at this patch?

I took a look at it when you sent it out, and found it so obviously and trivially correct that I expected that others will soon say it looked obviously the right thing to do. So I decided to wait until that to happen before applying it.

But nobody said anything, and forgot about it.
Let's see if it happens soon enough this time ;-).  Thanks for a reminder.
Christian Couder· Apr 27, 2011, 08:15 UTC · re: Junio C Hamano · lore

Re: [PATCH] git-rebase--interactive.sh: preserve-merges fails on merges created with no-ff

On Wed, Apr 27, 2011 at 7:00 AM, Junio C Hamano <gitster@pobox.com> wrote:
Show 7 quoted lines
> Andrew Wong <andrew.w@sohovfx.com> writes:
>
>> Could someone please take a look at this patch?
>
> I took a look at it when you sent it out, and found it so obviously and
> trivially correct that I expected that others will soon say it looked
> obviously the right thing to do.
Me too!
Show 6 quoted lines
>  So I decided to wait until that to
> happen before applying it.
>
> But nobody said anything, and forgot about it.
>
> Let's see if it happens soon enough this time ;-).  Thanks for a reminder.

Thanks from me too, Christian.

Junio C Hamano· Apr 27, 2011, 17:29 UTC · re: Andrew Wong · lore

Re: [PATCH] git-rebase--interactive.sh: preserve-merges fails on merges created with no-ff

Andrew Wong <andrew.w@sohovfx.com> writes:
> Could someone please take a look at this patch?
Can you add a test for this change, perhaps to either t3409 or t3414?
Andrew Wong· Apr 28, 2011, 04:35 UTC · re: Junio C Hamano · lore

[PATCH v2] git-rebase--interactive.sh: preserve-merges fails on merges created with no-ff

'git rebase' uses 'git merge' to preserve merges (-p). This preserves the original merge commit correctly, except when the original merge commit was created by 'git merge --no-ff'. In this case, 'git rebase' will fail to preserve the merge, because during 'git rebase', 'git merge' will simply fast-forward and skip the commit. For example:

               B
              / \
             A---M
            /
    ---o---O---P---Q
If we try to rebase M onto P, we lose the merge commit and this happens:
                 A---B
                /
    ---o---O---P---Q

To correct this, we simply do a "no fast-forward" on all merge commits when rebasing. Since by the time we decided to do a 'git merge' inside 'git rebase', it means there was a merge originally, so 'git merge' should always create a merge commit regardless of what the merge branches look like. This way, when rebase M onto P from the above example, we get:

                   B
                  / \
                 A---M
                /
    ---o---O---P---Q
Signed-off-by: Andrew Wong <andrew.kw.w@gmail.com>
---
 git-rebase--interactive.sh        |    2 +-
 t/t3409-rebase-preserve-merges.sh |   32 +++++++++++++++++++++++++++++++-
 2 files changed, 32 insertions(+), 2 deletions(-)
Show changes to 2 files +32 −2

git-rebase--interactive.sh, t/t3409-rebase-preserve-merges.sh

diff --git a/git-rebase--interactive.sh b/git-rebase--interactive.sh
index 5873ba4..c308529 100755
--- a/git-rebase--interactive.sh
+++ b/git-rebase--interactive.sh
@@ -339,7 +339,7 @@ pick_one_preserving_merges () {
 			# No point in merging the first parent, that's HEAD
 			new_parents=${new_parents# $first_parent}
 			if ! do_with_author output \
-				git merge $STRATEGY -m "$msg" $new_parents
+				git merge --no-ff $STRATEGY -m "$msg" $new_parents
 			then
 				printf "%s\n" "$msg" > "$GIT_DIR"/MERGE_MSG
 				die_with_patch $sha1 "Error redoing merge $sha1"
diff --git a/t/t3409-rebase-preserve-merges.sh b/t/t3409-rebase-preserve-merges.sh
index 19341e5..08201e2 100755
--- a/t/t3409-rebase-preserve-merges.sh
+++ b/t/t3409-rebase-preserve-merges.sh
@@ -27,7 +27,17 @@ export GIT_AUTHOR_EMAIL
 #    \
 #     B2       <-- origin/topic
 #
 # In both cases, 'topic' is rebased onto 'origin/topic'.
 
 test_expect_success 'setup for merge-preserving rebase' \
 	'echo First > A &&
@@ -61,6 +71,16 @@ test_expect_success 'setup for merge-preserving rebase' \
 		git commit -m "Merge origin/master into topic"
 	) &&
 
+	git clone ./. clone3 &&
+	(
+		cd clone3 &&
+		git checkout -b topic2 origin/topic &&
+		echo Sixth > A &&
+		git commit -a -m "Modify A3" &&
+		git checkout -b topic origin/topic &&
+		git merge --no-ff topic2
+	) &&
+
 	git checkout topic &&
 	echo Fourth >> B &&
 	git commit -a -m "Modify B2"
@@ -93,4 +113,14 @@ test_expect_success '--continue works after a conflict' '
 	)
 '
 
+test_expect_success 'rebase -p preserves no-ff merges' '
+	(
+	cd clone3 &&
+	git fetch &&
+	git rebase -p origin/topic &&
+	test 3 = $(git rev-list --all --pretty=oneline | grep "Modify A" | wc -l) &&
+	test 1 = $(git rev-list --all --pretty=oneline | grep "Merge branch" | wc -l)
+	)
+'
+
 test_done
@@ -27,7 +27,17 @@ export GIT_AUTHOR_EMAIL
 #     B2       <-- origin/topic
 #
-# In both cases, 'topic' is rebased onto 'origin/topic'.
+# Clone 3 (no-ff merge):
+#
+# A1--A2--B3   <-- origin/master
+#  \
+#   B1------M  <-- topic
+#    \     /
+#     \--A3    <-- topic2
+#      \
+#       B2     <-- origin/topic
+#
+# In all cases, 'topic' is rebased onto 'origin/topic'.
 
 test_expect_success 'setup for merge-preserving rebase' \
-- 
1.7.2.2

← back to recent threads